@@ -4,15 +4,12 @@
  * included with libxml++ as the file COPYING.
  */
 
-// Include glibmm/threads.h first. It must be the first file to include glib.h,
-// because it temporarily undefines G_DISABLE_DEPRECATED while it includes glib.h.
-#include <glibmm/threads.h> // For Glib::Threads::Mutex. Needed until the next API/ABI break.
-
 #include "libxml++/exceptions/wrapped_exception.h"
 #include "libxml++/parsers/parser.h"
 
 #include <libxml/parser.h>
 
+#include <mutex>
 #include <memory> //For unique_ptr.
 #include <map>
 
@@ -20,6 +17,9 @@
 
 namespace // anonymous
 {
+//TODO: When we can break ABI, remove ExtraParserData::xinclude_options_
+// and move all XInclude stuff to DomParser.
+
 // These are new data members that can't be added to xmlpp::Parser now,
 // because it would break ABI.
 struct ExtraParserData
@@ -27,7 +27,8 @@ struct ExtraParserData
   // Strange default values for throw_*_messages chosen for backward compatibility.
   ExtraParserData()
   : throw_parser_messages_(false), throw_validity_messages_(true),
-  include_default_attributes_(false), set_options_(0), clear_options_(0)
+  include_default_attributes_(false), set_options_(0), clear_options_(0),
+  xinclude_options_(0)
   {}
 
   Glib::ustring parser_error_;
@@ -37,23 +38,24 @@ struct ExtraParserData
   bool include_default_attributes_;
   int set_options_;
   int clear_options_;
+  int xinclude_options_;
 };
 
 std::map<const xmlpp::Parser*, ExtraParserData> extra_parser_data;
 // Different Parser instances may run in different threads.
 // Accesses to extra_parser_data must be thread-safe.
-Glib::Threads::Mutex extra_parser_data_mutex;
+std::mutex extra_parser_data_mutex;
 
 void on_parser_error(const xmlpp::Parser* parser, const Glib::ustring& message)
 {
-  Glib::Threads::Mutex::Lock lock(extra_parser_data_mutex);
+  std::lock_guard<std::mutex> lock(extra_parser_data_mutex);
   //Throw an exception later when the whole message has been received:
   extra_parser_data[parser].parser_error_ += message;
 }
 
 void on_parser_warning(const xmlpp::Parser* parser, const Glib::ustring& message)
 {
-  Glib::Threads::Mutex::Lock lock(extra_parser_data_mutex);
+  std::lock_guard<std::mutex> lock(extra_parser_data_mutex);
   //Throw an exception later when the whole message has been received:
   extra_parser_data[parser].parser_warning_ += message;
 }
@@ -71,7 +73,7 @@ Parser::~Parser()
 {
   release_underlying();
   delete exception_;
-  Glib::Threads::Mutex::Lock lock(extra_parser_data_mutex);
+  std::lock_guard<std::mutex> lock(extra_parser_data_mutex);
   extra_parser_data.erase(this);
 }
 
@@ -97,46 +99,58 @@ bool Parser::get_substitute_entities() const
 
 void Parser::set_throw_messages(bool val)
 {
-  Glib::Threads::Mutex::Lock lock(extra_parser_data_mutex);
+  std::lock_guard<std::mutex> lock(extra_parser_data_mutex);
   extra_parser_data[this].throw_parser_messages_ = val;
   extra_parser_data[this].throw_validity_messages_ = val;
 }
 
 bool Parser::get_throw_messages() const
 {
-  Glib::Threads::Mutex::Lock lock(extra_parser_data_mutex);
+  std::lock_guard<std::mutex> lock(extra_parser_data_mutex);
   return extra_parser_data[this].throw_parser_messages_;
 }
 
 void Parser::set_include_default_attributes(bool val)
 {
-  Glib::Threads::Mutex::Lock lock(extra_parser_data_mutex);
+  std::lock_guard<std::mutex> lock(extra_parser_data_mutex);
   extra_parser_data[this].include_default_attributes_ = val;
 }
 
 bool Parser::get_include_default_attributes()
 {
-  Glib::Threads::Mutex::Lock lock(extra_parser_data_mutex);
+  std::lock_guard<std::mutex> lock(extra_parser_data_mutex);
   return extra_parser_data[this].include_default_attributes_;
 }
 
 void Parser::set_parser_options(int set_options, int clear_options)
 {
-  Glib::Threads::Mutex::Lock lock(extra_parser_data_mutex);
+  std::lock_guard<std::mutex> lock(extra_parser_data_mutex);
   extra_parser_data[this].set_options_ = set_options;
   extra_parser_data[this].clear_options_ = clear_options;
 }
 
 void Parser::get_parser_options(int& set_options, int& clear_options)
 {
-  Glib::Threads::Mutex::Lock lock(extra_parser_data_mutex);
+  std::lock_guard<std::mutex> lock(extra_parser_data_mutex);
   set_options = extra_parser_data[this].set_options_;
   clear_options = extra_parser_data[this].clear_options_;
 }
 
+void Parser::set_xinclude_options_internal(int xinclude_options) noexcept
+{
+  std::lock_guard<std::mutex> lock(extra_parser_data_mutex);
+  extra_parser_data[this].xinclude_options_ = xinclude_options;
+}
+
+int Parser::get_xinclude_options_internal() const noexcept
+{
+  std::lock_guard<std::mutex> lock(extra_parser_data_mutex);
+  return extra_parser_data[this].xinclude_options_;
+}
+
 void Parser::initialize_context()
 {
-  Glib::Threads::Mutex::Lock lock(extra_parser_data_mutex);
+  std::unique_lock<std::mutex> lock(extra_parser_data_mutex);
 
   //Clear these temporary buffers:
   extra_parser_data[this].parser_error_.erase();
@@ -147,7 +161,7 @@ void Parser::initialize_context()
   // Take a copy of the extra data, so we don't have to access
   // the extra_parser_data map more than necessary.
   const auto extra_parser_data_this = extra_parser_data[this];
-  lock.release();
+  lock.unlock();
 
   //Disactivate any non-standards-compliant libxml1 features.
   //These are disactivated by default, but if we don't deactivate them for each context
@@ -232,7 +246,7 @@ void Parser::check_for_validity_messages() // Also checks parser messages
   bool parser_msg = false;
   bool validity_msg = false;
 
-  Glib::Threads::Mutex::Lock lock(extra_parser_data_mutex);
+  std::lock_guard<std::mutex> lock(extra_parser_data_mutex);
   if (!extra_parser_data[this].parser_error_.empty())
   {
     parser_msg = true;
@@ -347,13 +361,9 @@ void Parser::callback_error_or_warning(MsgType msg_type, void* ctx,
             break;
         }
       }
-      catch(const exception& e)
+      catch (...)
       {
-        parser->handleException(e);
-      }
-      catch(...)
-      {
-        parser->handleException(wrapped_exception(std::current_exception()));
+        parser->handle_exception();
       }
     }
   }
@@ -370,6 +380,42 @@ void Parser::handleException(const exception& e)
   //release_underlying();
 }
 
+void Parser::handle_exception()
+{
+  delete exception_;
+  exception_ = nullptr;
+
+  try
+  {
+    throw; // Rethrow current exception
+  }
+  catch (const exception& e)
+  {
+    exception_ = e.Clone();
+  }
+#ifdef LIBXMLXX_HAVE_EXCEPTION_PTR
+  catch (...)
+  {
+    exception_ = new wrapped_exception(std::current_exception());
+  }
+#else
+  catch (const std::exception& e)
+  {
+    exception_ = new exception(e.what());
+  }
+  catch (...)
+  {
+    exception_ = new exception("An exception was thrown that is not derived from std::exception or xmlpp::exception.\n"
+      "It could not be caught and rethrown because this platform does not support std::exception_ptr.");
+  }
+#endif
+
+  if (context_)
+    xmlStopParser(context_);
+
+  //release_underlying();
+}
+
 void Parser::check_for_exception()
 {
   check_for_validity_messages();
